#= require slender_data/queries/multi_query
#= require slender_data/queries/single_query
#= require slender_data/parameterizable


# High-level SDK for CRUD operations against a BREST API.
#
# Takes arguments as Queries (SingleQuery or MultiQuery),
# and expects the server to return self-describing data in raw query form,
# i.e. as "user: {...}" instead of just "{...}" when returning a user object.
# The server responses are forwarded as-is to the callers.
class slender_data.BrestLoader extends slender_data.Parameterizable

  constructor: (params) ->
    super params

    # The base url of the BREST API on the server.
    #
    # Example: provide '/api' to access users at '/api/users'.
    @base_url = params.base_url or ''


  # Creates the entry described by the given query on the server.
  create: (entry, callback) ->
    data = {}
    data[@entry_name] = entry

    jQuery.ajax
      url: @entries_url()
      type: 'POST'
      data: data
      success: callback


  # Creates the entries described by the given multi-query on the server.
  create_many: (entries, callback) ->
    data = {}
    data[@entries_name] = entries

    jQuery.ajax
      url: @entries_url()
      type: 'POST'
      data: data
      success: callback


  # Deletes the entry described by the given query on the server.
  delete: (query, callback) ->
    query = @_querify_single query
    jQuery.ajax
      url: @entry_url(query)
      type: 'DELETE'
      success: callback


  # Deletes the entries described by the given multi-query on the server.
  delete_many: (multi_query, callback) ->
    multi_query = @_querify_multi multi_query
    data = {}
    data[@ids_name] = multi_query.ids()
    jQuery.ajax
      url: @entries_url(data)
      type: 'DELETE'
      success: callback


  # Returns the url to access a single entry.
  entry_url: (query) ->
    query = @_querify_single query
    "#{@base_url}/#{@entries_name}/#{query.id()}.json"


  # Returns the url to access the collection of entries.
  entries_url: (params) ->
    data = if params then "?#{@url_data_encode params}" else ""
    "#{@base_url}/#{@entries_name}.json#{data}"


  # Merges the entries described by the given ids
  # into a new entry with the given data.
  merge: (old_entry_ids, new_entry, callback) ->
    data =
      old_snippet_ids: old_entry_ids
      merged_snippet_data: new_entry
    $.ajax
      url: "#{@base_url}/#{@entries_name}/merge.json"
      type: 'POST'
      data: data
      success: callback

  # Fetches the entry with the given key from the server.
  read: (query, callback) ->
    query = @_querify_single query
    jQuery.get @entry_url(query), callback


  # Loads all objects from the server.
  # Provides the given params as parameters to the GET request.
  read_many: (multi_query, callback) ->
    multi_query = @_querify_multi multi_query
    data = {}
    data[multi_query.ids_name] = multi_query.ids().join(' ')
    jQuery.ajax
      url: @entries_url()
      type: 'GET'
      cache: no
      data: data
      success: callback


  # Updates the given entry on the server.
  #
  # The entry can be a partial diff of the real entry,
  # but must contain the key column.
  update: (entry, callback) ->
    data = {}
    data[@entry_name] = entry

    jQuery.ajax
      url: @entry_url(entry: entry)
      type: 'PUT'
      data: data
      success: callback


  # Updates the given entries on the server.
  #
  # The entry can be a partial diff of the real entry,
  # but must contain the key column.
  update_many: (entries, callback) ->
    data = {}
    data[@entries_name] = entries

    jQuery.ajax
      url: @entries_url()
      type: 'PUT'
      data: data
      success: callback


  # URL-encodes the given JSON data.
  url_data_encode: (params) ->
    data = []
    for own key, value of params
      if _.isArray(value)
        value = value.join '+'
      data.push "#{key}=#{value}"
    data = data.join '&'

