#= require slender_data/query_cache
#= require slender_data/queries/single_query
#= require slender_data/queries/multi_query
#= require slender_data/parameterizable


# Stores JSON data objects, and tracks changes to them.
class slender_data.EntryStore extends slender_data.Parameterizable

  constructor: (params) ->
    super params

    # Unmodified copies of the serverside data.
    @server_data = new slender_data.QueryCache params

    # Replicas of the server data, to be modified by the client.
    @client_data = new slender_data.QueryCache params


  # Adds the given entry to the caches.
  # Treats the given entry as coming from the server.
  add: (entry, return_editable_version = no) ->
    @server_data.add entry
    result = undefined
    if return_editable_version
      result = @client_data.add_clone entry
    else
      @client_data.remove entry
    result


  # Adds the given data objects to the server cache.
  add_many: (entries, return_editable_version = no) ->
    result = (@add(entry, return_editable_version) for entry in entries)
    result if return_editable_version


  # Removes the entry corresponding to the given query
  # from this DataStore.
  delete: (entry) ->
    @client_data.remove entry
    @server_data.remove entry


  # Removes the entries corresponding to the given multi-query
  # from this DataStore.
  delete_many: (entries) ->
    @client_data.remove_many entries
    @server_data.remove_many entries


  # Returns an object that contains only the changed attributes
  # of the entry destribed by the given query.
  # Returns false if the entry is not changed.
  entry_changes: (entry) ->
    server_version = @server_data.get entry: entry
    changes = slender_data.object_diff server_version, entry
    if slender_data.object_length(changes) > 0
      changes[@id_name] = entry[@id_name]
      changes
    else
      false


  # Returns an array containing only the changed objects.
  # Each object contains only the changed columns + key column.
  entries_changes: (entries) ->
    result = []
    for entry in entries
      if (entry_changes = @entry_changes entry)
        result.push entry_changes
    result


  # Returns the entry described by the given query,
  # or undefined if it does not exist in any of the caches.
  get: (query) ->
    query = @_querify_single query

    # Try to use client_data cache.
    unless (result = @client_data.get query)

      # No data in client cache --> try to use server cache.
      if (server_entry = @server_data.get query)
        result = @client_data.add_clone server_entry

    result


  # Returns the entries described by the given multi-query,
  # as well as a list of entries not found.
  get_many: (multi_query) ->
    found = []

    # Load client-side data.
    client_results = @client_data.get_many @_querify_multi(multi_query)
    $.merge found, client_results.found
    missing_query = client_results.missing

    if not missing_query.empty()
      # Some entries not in client cache --> look in server cache.
      server_results = @server_data.get_many missing_query
      $.merge found, @client_data.add_clones(server_results.found)
      missing_query = server_results.missing

    found: found, missing: missing_query

