#= require slender_data/queries/single_query
#= require slender_data/queries/multi_query


# Tracks which entries are currently getting loaded,
# and manages requests for entry data.
class slender_data.LoadingTracker extends slender_data.Parameterizable

  constructor: (params) ->
    super params

    # Maps currently loading entries to callbacks waiting for their data.
    # Key = entry id, value = array of callbacks
    @loading_entries = {}


  # Returns whether the entry described by the given query
  # is currently loading.
  is_loading: (query) ->
    query = @_querify_single query
    console.log "LoadingTracker.is_loading: entry #{query.id()} loading status: #{@loading_entries[query.id()]?}" if window.JS_TRACING
    @loading_entries[query.id()]?


  # Notifies this LoadingTracker that the given entry has arrived.
  entry_arrived: (entry) ->
    console.log "LoadingTracker: received entry #{entry.id}" if window.JS_TRACING

    # Call the callbacks registered for the loaded entry.
    if (callbacks = @loading_entries[entry[@id_name]])
      callback(entry) for callback in callbacks

    # Remove the id of the loaded entry from the loading list.
    delete @loading_entries[entry[@id_name]]


  # Notifies this LoadingTracker that the given entries have arrived.
  entries_arrived: (entries) ->
    console.log "LoadingTracker: received entries #{(s.id for s in entries).join '+'}" if window.JS_TRACING
    @entry_arrived(entry) for entry in entries
    return


  # Registers the given entry as currently loading from the server.
  start_loading: (query) ->
    query = @_querify_single query
    console.log "LoadingTracker: starting to load #{query.id()}" if window.JS_TRACING
    @loading_entries[query.id()] ?= []


  # Registers the given entries as currently loading from the server.
  start_loadings: (multi_query) ->
    multi_query = @_querify_multi multi_query
    console.log "LoadingTracker: starting to load multiple #{multi_query.ids()}" if window.JS_TRACING
    @start_loading(query) for query in multi_query.transpose()
    return


  # Calls the given callback when the entry described by the given query
  # is loaded.
  # Returns immediately with the given query when the entry is not loading.
  wait_for_entry_loaded: (query, callback) ->
    query = @_querify_single query

    if @is_loading(query)
      console.log "LoadingTracker.wait_for_entry_loaded: entry #{query.id()} is loading, waiting for it" if window.JS_TRACING
      @loading_entries[query.id()].push callback
      return undefined
    else
      console.log "LoadingTracker.wait_for_entry_loaded: entry #{query.id()} is not loading" if window.JS_TRACING
      return query


  # Calls the given callback when all currently loading entries described by
  # the given multi-query have finished loading.
  # Returns a multi-query describing the not currently loading entries.
  wait_for_entries_loaded: (multi_query, callback) ->
    multi_query = @_querify_multi multi_query
    result = []
    asynchronizer = new Asynchronizer multi_query.ids()
    for query in multi_query.transpose()
      not_loading_entry_query = @wait_for_entry_loaded query, (loaded_entry) =>
        asynchronizer.check loaded_entry[@id_name], loaded_entry
      if not_loading_entry_query
        # The current entry is not loading --> don't wait for it
        # to finish loading, and add it to the result.
        asynchronizer.check not_loading_entry_query.id()
        result.push not_loading_entry_query.selector_value()
    asynchronizer.then callback
    multi_query.replicate result
