#= require slender_data/brest_loader
#= require slender_data/entry_store
#= require slender_data/loading_tracker
#= require slender_data/query_cache
#= require slender_data/parameterizable
#= require slender_data/tools

# Provides server-backed persistence services for JSON data models.
#
# This class operates against a BRESTful data API.
#
# Operations that act directly on entries, and are always given entries
# (add_to_cache, create, update) are given the entries directly.
# Operations that need to look up entries before doing something with them,
# and don't necessarily need the full entry to do that, (delete, get, load)
# are given queries.
class slender_data.PersistenceManager extends slender_data.Parameterizable

  constructor: (params) ->
    super params

    # Stores the data objects here on the client,
    # and tracks changes that are made to them here.
    @entry_store = new slender_data.EntryStore params

    # Tracks which entries are currently loading.
    @loading_tracker = new slender_data.LoadingTracker params

    # Loads/stores data to/from the server.
    @server_api = new slender_data.BrestLoader params


  # Adds the given serverside entry to the entry store.
  # Does not update the server.
  # Does not provide an editable version of the data.
  # Call "get" to get one.
  add_to_cache: (entry, return_editable_version = no) ->
    @entry_store.add entry, return_editable_version


  # Adds the given serverside entries to the entry store.
  add_many_to_cache: (entries, return_editable_version = no) ->
    @entry_store.add_many entries, return_editable_version


  # Creates the described object on the server.
  create: (entry, callback) ->
    @server_api.create entry, (server_response) =>
      server_entry = server_response[@entry_name]
      @add_to_cache server_entry
      callback(@get entry: server_entry) if callback

  # Creates the described objects on the server.
  #
  # Returns nothing. Call "get_many" to get editable copies.
  create_many: (entries, callback) ->
    @server_api.create_many entries, (server_response) =>
      @add_many_to_cache server_response[@entries_name]
      callback() if callback


  # Deletes the described entry from the server.
  delete: (query, callback) ->
    query = query
    @server_api.delete query, (server_response) =>
      @entry_store.delete server_response
      callback(server_response) if callback


  # Deletes the described entries from the server.
  delete_many: (multi_query, callback) ->
    @server_api.delete_many multi_query, (server_response) =>
      @entry_store.delete_many server_response
      callback(server_response) if callback


  # Synchronously returns the described entry from the cache,
  # or undefined if there is no cached entry.
  # Does not query the server.
  get: (query) =>
    @entry_store.get query


  # Synchronously returns all described entries that are found in the cache,
  # and a MultiQuery for the entries that are not there.
  get_many: (multi_query) ->
    @entry_store.get_many multi_query


  # Returns the described entry from any available data source.
  load: (query, callback) ->
    query = @_querify_single query
    if (entry = @get query)
      return callback entry

    # The entry was not found in the cache --> see if it is currently loading.
    entry_loading = !@loading_tracker.wait_for_entry_loaded query, (server_entry) =>
      callback @add_to_cache(server_entry, yes)
    return if entry_loading

    # The entry is not currently loading --> load it from the server.
    @server_api.read query, (server_response) =>
      callback @add_to_cache server_response[@entry_name], yes


  # Loads all described entries from any available data source.
  load_many: (multi_query, callback) ->
    multi_query = @_querify_multi multi_query
    console.log "PersistenceManager.load_many: loading #{multi_query.ids().join '+'}" if window.JS_TRACING
    result = []

    cache_result = @get_many multi_query
    console.log "PersistenceManager.load_many: found '#{(entry.id for entry in cache_result.found).join '+'}' in cache." if window.JS_TRACING
    result = _(result).union cache_result.found
    if cache_result.missing.empty()
      # All requested entries found in cache --> return them immediately.
      return callback result if callback

    # Some entries were not in the cache.
    asynchronizer = new Asynchronizer ['loading_tracker', 'server_api']

    # Use the currently loading entries.
    loading_tracker_missing_multiquery = @loading_tracker.wait_for_entries_loaded cache_result.missing, (loading_tracker_entries) =>
      client_entries = @add_many_to_cache loading_tracker_entries, yes
      asynchronizer.check 'loading_tracker', client_entries

    if loading_tracker_missing_multiquery.empty()
      # Nothing needs to be loaded from the server --> check the server part off now.
      asynchronizer.check 'server_api'
    else
      # Some entries are not loading --> load them from the server.
      @loading_tracker.start_loadings loading_tracker_missing_multiquery
      @server_api.read_many loading_tracker_missing_multiquery, (server_response) =>
        client_entries = @add_many_to_cache server_response[@entries_name], yes
        @loading_tracker.entries_arrived server_response[@entries_name]
        asynchronizer.check 'server_api', client_entries

    asynchronizer.then (entries) =>
      # All entries have been loaded from all the different sources.
      callback _.chain(result).union(entries).flatten().compact().value() if callback


  # Merges the entries described by the given ids
  # into a new entry with the given data.
  merge: (old_entry_ids, new_entry_data, callback) ->
    @server_api.merge old_entry_ids, new_entry_data, (server_response) =>
      @entry_store.delete_many ids: old_entry_ids
      callback @entry_store.add server_response[@entry_name], yes


  # Updates the given object.
  #
  # The given object must already exist on the server,
  # and have a proper value in the id attribute.
  update: (entry, callback) ->
    if (changes = @entry_store.entry_changes entry)
      # There are changes --> update the server.
      @server_api.update changes, (server_response) =>
        @entry_store.add server_response[@entry_name]
        callback(@get entry: server_response[@entry_name]) if callback
    else
      # No changes --> call the callback directly.
      callback(entry) if callback


  # Bulk-updates the given objects.
  # All the existing objects must already exist on the server,
  # and have a value in the id attribute.
  update_many: (entries, callback) ->
    if (changes = @entry_store.entries_changes entries).length > 0
      # There are changes --> update the server.
      @server_api.update_many changes, (server_response) =>
        @entry_store.add_many server_response[@entries_name]
        callback(server_response) if callback
    else
      callback_data = {}
      callback_data[@entries_name] = entries
      callback(callback_data) if callback

